/***************************************************************************
                          fractaldisplay.cpp  -  description
                             -------------------
    begin                : Sam Mai 10 2003
    copyright            : (C) 2003 by Thomas Dreibholz
    email                : dreibh@exp-math.uni-essen.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "fractaldisplay.h"


#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define MAX(a,b) (((a) > (b)) ? (a) : (b))


FractalDisplay::FractalDisplay(QWidget *parent, const char *name ) : QWidget(parent,name)
{
   Image = new QImage();
   Q_CHECK_PTR(Image);
   OffsetX = 0;
   OffsetY = 0;
   MarkX1  = 0;
   MarkY1  = 0;
   MarkX2  = 0;
   MarkY2  = 0;
   Marking = false;
}

FractalDisplay::~FractalDisplay()
{
   delete Image;
}

bool FractalDisplay::reset(const unsigned int width, const unsigned int height)
{
   if(Image) {
     delete Image;
   }
   Image = new QImage(width, height, 32);
   Q_CHECK_PTR(Image);
   for(int y = 0;y < Image->height();y++) {
      for(int x = 0;x < Image->width();x++) {
         Image->setPixel(x, y, qRgb(255, 255, 255));
      }
   }

   OffsetX = 0;
   OffsetY = 0;
   MarkX1  = 0;
   MarkY1  = 0;
   MarkX2  = 0;
   MarkY2  = 0;
   Marking = false;
   return(true);
}

void FractalDisplay::resizeEvent(QResizeEvent* resizeEvent)
{
   update();
}

void FractalDisplay::paintEvent(QPaintEvent* paintEvent)
{
   QPainter p;
   p.begin(this);
   p.drawImage(paintEvent->rect().left(), paintEvent->rect().top(),
               *Image,
               paintEvent->rect().left() + OffsetX,
               paintEvent->rect().top()  + OffsetY,
               paintEvent->rect().width(),
               paintEvent->rect().height());
   if(Marking) {
      drawMarkerRect(&p, OldMarkX1, OldMarkY1, OldMarkX2, OldMarkY2);
   }
   p.end();
}

void FractalDisplay::getMarkPosition(QMouseEvent* mouseEvent, int& x, int& y)
{
   x = mouseEvent->x() + (int)OffsetX;
   if(x < 0) {
      x = 0;
   }
   if(x >= Image->width()) {
      x = Image->width() - 1;
   }
   y = mouseEvent->y() + (int)OffsetY;
   if(y < 0) {
      y = 0;
   }
   if(y >= Image->height()) {
      y = Image->height() - 1;
   }
}

void FractalDisplay::mousePressEvent(QMouseEvent* mouseEvent)
{
   if(mouseEvent->button() & LeftButton) {
      getMarkPosition(mouseEvent, MarkX1, MarkY1);
      MarkX2 = MarkX1;
      MarkY2 = MarkY1;
      showMarker();
   }
}

void FractalDisplay::mouseReleaseEvent(QMouseEvent* mouseEvent)
{
   LastOffsetUpdate = QTime();
   mouseMoveEvent(mouseEvent);

   if((mouseEvent->button() & LeftButton) && (Marking)) {
      getMarkPosition(mouseEvent, MarkX2, MarkY2);
      showMarker();
      if((MarkX1 != MarkX2) && (MarkY1 != MarkY2)) {
         selection(MarkX1, MarkY1, MarkX2, MarkY2);
      }
   }
   else if((mouseEvent->button() & RightButton) && (Marking)) {
      QPainter p;
      p.begin(this);
      drawMarkerRect(&p, OldMarkX1, OldMarkY1, OldMarkX2, OldMarkY2, false);
      p.end();
      Marking = false;
      selection(0, 0, 0, 0);
   }
   if(mouseEvent->button() & MidButton) {
      zoom();
   }
}

void FractalDisplay::mouseMoveEvent(QMouseEvent* mouseEvent)
{
   if((Marking) && (LastOffsetUpdate.elapsed() >= 50)) {
      int movex = 0;
      int movey = 0;
      if(mouseEvent->x() < 0) {
         movex = MAX(mouseEvent->x(), -((int)OffsetX));
      }
      else if(mouseEvent->x() >= width()) {
         movex = MIN(mouseEvent->x() - width(), Image->width() - width() - (int)OffsetX);
      }
      if(mouseEvent->y() < 0) {
         movey = MAX(mouseEvent->y(), -((int)OffsetY));
      }
      else if(mouseEvent->y() >= height()) {
         movey = MIN(mouseEvent->y() - height(), Image->height() - height() - (int)OffsetY);
      }

      if((movex != 0) || (movey != 0)) {
         offsetUpdate((int)OffsetX + movex, (int)OffsetY + movey);
         LastOffsetUpdate = QTime::currentTime();
      }

      getMarkPosition(mouseEvent, MarkX2, MarkY2);      
      showMarker();
   }
}

void FractalDisplay::drawMarkerRect(QPainter* painter, int x1, int y1, int x2, int y2, bool draw)
{
   const int x  = MIN(x1, x2) - OffsetX;
   const int y  = MIN(y1, y2) - OffsetY;
   const int w  = (MAX(x1, x2) - OffsetX) - x;
   const int h  = (MAX(y1, y2) - OffsetY) - y;
   const int bw = MIN(w / 2, 5);
   const int bh = MIN(h / 2, 5);

   if(draw) {
      painter->fillRect(x - bw, y - bh,    w + 2 * bw, bh,     Qt::blue);
      painter->fillRect(x - bw, y + h + 1, w + 2 * bw, bh,     Qt::blue);
      painter->fillRect(x - bw, y,         bw,         h + 1,  Qt::blue);
      painter->fillRect(x + w,  y,         bw,         h + 1,  Qt::blue);
   }
   else {
      painter->drawImage(x - bw, y - bh,    *Image, x - bw + (int)OffsetX, y - bh +    (int)OffsetY, w + 2 * bw,  bh);
      painter->drawImage(x - bw, y + h + 1, *Image, x - bw + (int)OffsetX, y + h + 1 + (int)OffsetY, w + 2 * bw,  bh);
      painter->drawImage(x - bw, y,         *Image, x - bw + (int)OffsetX, y +         (int)OffsetY, bw, h + 1);
      painter->drawImage(x + w,  y,         *Image, x + w  + (int)OffsetX, y +         (int)OffsetY, bw, h + 1);

   }
}

void FractalDisplay::showMarker()
{
   QPainter p;
   p.begin(this); 
   if(Marking) {
      drawMarkerRect(&p, OldMarkX1, OldMarkY1, OldMarkX2, OldMarkY2, false);
   }
   drawMarkerRect(&p, MarkX1, MarkY1, MarkX2, MarkY2);
   Marking = true;
   OldMarkX1 = MarkX1;
   OldMarkY1 = MarkY1;
   OldMarkX2 = MarkX2;
   OldMarkY2 = MarkY2;
   p.end();
}
