/***************************************************************************
                          fractalgeneratorview.cpp  -  description
                             -------------------
    begin                : Sam Mai 10 10:39:46 CEST 2003
    copyright            : (C) 2003 by Thomas Dreibholz
    email                : dreibh@exp-math.uni-essen.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// include files for Qt
#include <qpainter.h>
#include <qimage.h>
#include <qapplication.h>
#include <iostream>  // ???????

// application specific includes
#include "fractalgeneratorview.h"
#include "fractalgeneratorview.moc"
#include "fractalgeneratordoc.h"
#include "fractalgenerator.h"

#include "mandelbrot.h" // ?????


#define MIN(a,b) ((a < b) ? a : b)
#define MAX(a,b) ((a > b) ? a : b)


FractalGeneratorView::FractalGeneratorView(QWidget *parent, const char *name) : QWidget(parent, name)
{
   theApp = static_cast<FractalGeneratorApp *>(parent);
   Thread = NULL;
   installEventFilter(this);
   Buffer = new FractalBuffer();
   Q_CHECK_PTR(Buffer);
   Display = new FractalDisplay(this);
   Q_CHECK_PTR(Display);
   Display->setBackgroundColor(black);
   connect(Display, SIGNAL(offsetUpdate(int, int)), this, SLOT(offsetUpdate(int, int)));
   connect(Display, SIGNAL(zoom()), this, SLOT(zoomIn()));
   connect(Display, SIGNAL(selection(unsigned int, unsigned int, unsigned int, unsigned int)),
           this, SLOT(selectionUpdate(unsigned int, unsigned int, unsigned int, unsigned int)));
   XScrollBar = new QScrollBar(0, 0, 1, 64, 0, Qt::Horizontal, this);
   Q_CHECK_PTR(XScrollBar);
   connect(XScrollBar, SIGNAL(valueChanged(int)), this, SLOT(xScrollBarChange(int)));
   YScrollBar = new QScrollBar(0, 0, 1, 64, 0, Qt::Vertical, this);
   Q_CHECK_PTR(YScrollBar);
   connect(YScrollBar, SIGNAL(valueChanged(int)), this, SLOT(yScrollBarChange(int)));
   ControlLED = new KLed(Qt::green, this);
   Q_CHECK_PTR(ControlLED);

   QGridLayout* layout = new QGridLayout(this, 2, 2);
   Q_CHECK_PTR(layout);
   layout->addWidget(Display, 0, 0);
   layout->addWidget(XScrollBar, 1, 0);
   layout->addWidget(YScrollBar, 0, 1);
   layout->addWidget(ControlLED, 1, 1);

   SizeX = 800;
   SizeY = 600;
   Display->reset(SizeX, SizeY);
   Buffer->reset(Display->imageWidth(), Display->imageHeight());

   Algorithm = FractalAlgorithmInterface::getAlgorithm(0);
   ColorScheme = ColorSchemeInterface::getColorScheme(0);
   C1 = Algorithm->defaultC1();
   C2 = Algorithm->defaultC2();
   Selection     = false;
   ProgStep      = 8;
   Algorithm->configure(Display->imageWidth(),
                        Display->imageHeight(),
                        C1, C2,
                        Algorithm->defaultMaxIterations());
   ColorScheme->configure(Algorithm->getMaxIterations());
   startCalculation();

   updateScrollBars();
}

FractalGeneratorView::~FractalGeneratorView()
{
   delete Buffer;
}

void FractalGeneratorView::xScrollBarChange(int value)
{
   int newOffset = value;
   
   if(newOffset + Display->width() >= (int)Display->imageWidth()) {
      newOffset = Display->imageWidth() - Display->width();
   }
   if(newOffset < 0) {
      newOffset = 0;
   }

   if(Display->offsetX() != (unsigned int)newOffset) {
      Display->setOffsetX((unsigned int)newOffset);
      Display->update();      
   }
}


void FractalGeneratorView::changeSize(int X, int Y)
{
   if(Thread != NULL) {
         stopCalculation();
	     }
   SizeX = X;
   SizeY = Y;
   Display->reset(SizeX, SizeY);
   Buffer->reset(Display->imageWidth(), Display->imageHeight());


   Selection     = false;
   Algorithm->changeSize(Display->imageWidth(),
                        Display->imageHeight());
   updateScrollBars();
}


void FractalGeneratorView::changeC1C2(std::complex<double> NewC1, std::complex<double> NewC2)
{
   if(Thread != NULL) {
         stopCalculation();
     }
   C1 = NewC1;
   C2 = NewC2;
}


void FractalGeneratorView::yScrollBarChange(int value)
{
   int newOffset = value;

   if(newOffset + Display->height() >= (int)Display->imageHeight()) {
      newOffset = Display->imageHeight() - Display->height();
   }
   if(newOffset < 0) {
      newOffset = 0;
   }

   if(Display->offsetY() != (unsigned int)newOffset) {
      Display->setOffsetY((unsigned int)newOffset);
      Display->update();
   }
}

void FractalGeneratorView::offsetUpdate(int newOffsetX, int newOffsetY)
{
   XScrollBar->setValue(newOffsetX);
   YScrollBar->setValue(newOffsetY);
}

void FractalGeneratorView::selectionUpdate(unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2)
{
   SelectionC1 = std::complex<double>(C1.real() + x1 * ((C2.real() - C1.real()) / Display->imageWidth()),
                                      C1.imag() + y1 * ((C2.imag() - C1.imag()) / Display->imageHeight()));
   SelectionC2 = std::complex<double>(C1.real() + x2 * ((C2.real() - C1.real()) / Display->imageWidth()),
                                      C1.imag() + y2 * ((C2.imag() - C1.imag()) / Display->imageHeight()));
   if((SelectionC1 != std::complex<double>(0.0, 0.0)) &&
      (SelectionC2 != std::complex<double>(0.0, 0.0))) {
      Selection = true;
   }
   else {
      Selection = false;
   }
}

void FractalGeneratorView::zoomIn()
{
   if(Selection) {
      if(Thread != NULL) {
         stopCalculation();
      }
      zoomList.push_back(std::pair<std::complex<double>, std::complex<double> >(C1, C2));
      const double halfwidth  = (SelectionC2.real() - SelectionC1.real()) / 2.0;
      const double halfheight = (SelectionC2.imag() - SelectionC1.imag()) / 2.0;
      const std::complex<double> center = std::complex<double>(SelectionC1.real() + halfwidth,
                                                               SelectionC1.imag() + halfheight);
      const double newhalfheight = halfwidth * ((double)Display->imageHeight() / (double)Display->imageWidth());
      C1 = std::complex<double>(center.real() - halfwidth, center.imag() + newhalfheight);
      C2 = std::complex<double>(center.real() + halfwidth, center.imag() - newhalfheight);
      Selection = false;
      Buffer->clear();
      Display->reset(Display->imageWidth(), Display->imageHeight());
      Algorithm->configure(Display->imageWidth(),
                           Display->imageHeight(),
                           C1, C2,
                           *Algorithm->getMaxIterations());
      startCalculation();
      theApp->updateZoomBackPossible();
   }
}

void FractalGeneratorView::updateScrollBars()
{
   XScrollBar->setMaxValue(MAX(0, Display->imageWidth()  - Display->width()));
   XScrollBar->setPageStep(MIN(Display->width(), Display->imageWidth()) / 4);
   YScrollBar->setMaxValue(MAX(0, Display->imageHeight() - Display->height()));
   YScrollBar->setPageStep(MIN(Display->height(), Display->imageHeight()) / 4);
}

FractalGeneratorDoc *FractalGeneratorView::getDocument() const
{
  FractalGeneratorApp *theApp=(FractalGeneratorApp *) parentWidget();

  return theApp->getDocument();
}

void FractalGeneratorView::print(KPrinter* printer)
{
   int width  = Display->imageWidth();
   int height = Display->imageHeight();
   char title[512];
   snprintf((char*)&title, sizeof(title), "%s  -  c1=%f - %fi; c2=%f - %fi; %d iterations",
            Algorithm->getName(),
            C1.real(), C1.imag(), C2.real(), C2.imag(), *Algorithm->getMaxIterations());

   QFont font("Times", 8);
   font.setBold(true);
   QFontMetrics fontMetrics(font);
   QRect        boundingRect = fontMetrics.boundingRect(title);
   int textwidth  = boundingRect.width();
   int textheight = boundingRect.height();
   
//   height += (height / 40) + 2;

   printer->setDocName(title);
   printer->setCreator("Thomas Dreibholz's Fractal Generator");


   int pagesx = 1;
   int pagesy = 1;

   QPainter p;
   p.begin(printer);
   const int xstep = width  / pagesx;
   const int ystep = height / pagesy;
   for(int y = 0;y < pagesy;y++) {
     for(int x = 0;x < pagesx;x++) {
        p.setWindow(0, 0, xstep, ystep);
        p.drawImage(0, 0, *(Display->image()), x * xstep, y * ystep, xstep, ystep);

        if((x == pagesx - 1) && (y == pagesy - 1)) {
           p.setWindow(0, 0, (int)(1.1 * textwidth), textheight * 50);
           p.setPen(Qt::black);
           p.setFont(font);
           p.drawText(0, textheight * 48, (int)(1.1 * textwidth), textheight, Qt::AlignHCenter|Qt::AlignVCenter, title);
        }
        else {
           printer->newPage();         
        }

        /*
        p.setWindow(0, 0, width, height);  
        p.setPen(Qt::red);
        p.drawLine(0, 0, Display->imageWidth() - 1, Display->imageHeight() - 1);
        p.drawLine(0, Display->imageHeight() - 1, Display->imageWidth() - 1, 0);
        p.setPen(Qt::black);
        char str[16];
        snprintf((char*)&str,sizeof(str),"x=%d y=%d\n",x,y);
        p.drawText(0, 0, str);
        p.drawRect(0, 0, Display->imageWidth() - 1, Display->imageHeight() - 1);
        */
     }
   }                                                              
   p.end();
}

void FractalGeneratorView::resizeEvent(QResizeEvent* resizeEvent)
{
   updateScrollBars();
}

void FractalGeneratorView::startCalculation()
{
   if(Thread == NULL) {
      Thread = new FractalCalculationThread(this, Algorithm, ColorScheme, Buffer, Display->image(), ProgStep);
      Q_CHECK_PTR(Thread);
      Thread->start();
      ControlLED->setColor(Qt::yellow);
   }
}

void FractalGeneratorView::stopCalculation()
{
   if(Thread != NULL) {
      Thread->stop();
      while(Thread != NULL) {
        qApp->processEvents();
      }
   }
}


bool FractalGeneratorView::eventFilter(QObject* object, QEvent* event)
{
   if(Thread) {
      if(event->type() == QEvent::User) {
         Display->update();
      }
      else if(event->type() == (QEvent::Type)(QEvent::User + 1)) {
         Display->update();
         Thread->stop();
         delete Thread;
         Thread = NULL;
         ControlLED->setColor(Qt::green);
      }
   }
   return(false);
}


void FractalGeneratorView::configChanged()
{
	//! \todo If user has zoomed in this has to be reset here
	if(Thread != NULL)
	{
		stopCalculation();
    }
	Buffer->clear();
    Display->reset(Display->imageWidth(), Display->imageHeight());
      
	startCalculation();
}


void FractalGeneratorView::changeAlgorithm(int index)
{
	if(Thread != NULL)
	{
		stopCalculation();
    }
	Algorithm = FractalAlgorithmInterface::getAlgorithm(index);
	C1 = Algorithm->defaultC1();
	C2 = Algorithm->defaultC2();
	Selection     = false;
	Algorithm->configure(Display->imageWidth(), Display->imageHeight(),
                        C1, C2,
                        Algorithm->defaultMaxIterations());
	ColorScheme->configure(Algorithm->getMaxIterations());
    zoomList.clear();
   
}

void FractalGeneratorView::changeColorScheme(int index)
{
	if(Thread != NULL)
	{
		stopCalculation();
    }
	ColorScheme = ColorSchemeInterface::getColorScheme(index);
	ColorScheme->configure(Algorithm->getMaxIterations());
   
}


void FractalGeneratorView::resetZoom()
{
	if(Thread != NULL)
	{
		stopCalculation();
    }
    C1 = Algorithm->defaultC1();
    C2 = Algorithm->defaultC2();
    Algorithm->configure(Display->imageWidth(),
                        Display->imageHeight(),
                        C1, C2,
                        *Algorithm->getMaxIterations());
    zoomList.clear();
}


void FractalGeneratorView::zoomBack()
{
if(Thread != NULL)
	{
		stopCalculation();
    }
    if(zoomList.size() > 0)
    {
        C1 = zoomList.back().first;
        C2 = zoomList.back().second;
        zoomList.pop_back();
        Algorithm->configure(Display->imageWidth(),
                        Display->imageHeight(),
                        C1, C2,
                        *Algorithm->getMaxIterations());
    }
    
}

bool FractalGeneratorView::isZoomBackPossible()
{
    return zoomList.size() > 0;
}
